"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var is_1 = require("./is");
var speculation_1 = require("./speculation");
var noop = function () {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
};
exports.typeDetector = function (type) { return function (value) {
    return value && is_1.isObject(value) && value.type === type;
}; };
exports.CALL = 'CALL';
function call(fn) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    return {
        type: exports.CALL,
        fn: fn,
        args: args,
    };
}
exports.call = call;
var isCall = exports.typeDetector(exports.CALL);
function callEffect(_a, promisify, cancel) {
    var _this = this;
    var fn = _a.fn, args = _a.args;
    return speculation_1.default(function (resolve, reject, onCancel) {
        var gen;
        if (Array.isArray(fn)) {
            var obj = fn[0], fnName = fn[1], fargs = fn.slice(2);
            gen = obj[fnName].apply(obj, fargs);
        }
        else {
            gen = fn.call.apply(fn, [_this].concat(args));
        }
        if (!gen || typeof gen.next !== 'function') {
            return resolve(gen);
        }
        promisify(gen, cancel)
            .then(resolve)
            .catch(reject);
        onCancel(function (msg) {
            if (typeof gen.next === 'function') {
                try {
                    gen.throw(msg);
                }
                catch (err) {
                    console.log(err);
                }
            }
            reject(msg);
        });
    }, cancel);
}
exports.ALL = 'ALL';
function all(effects) {
    return {
        type: exports.ALL,
        effects: effects,
    };
}
exports.all = all;
var isAll = exports.typeDetector(exports.ALL);
function allEffect(_a, promisify, cancel) {
    var effects = _a.effects;
    var ctx = this;
    if (Array.isArray(effects)) {
        var mapFn = function (effect) {
            return effectHandler.call(ctx, { effect: effect, promisify: promisify, cancel: cancel });
        };
        var eff_1 = effects.map(mapFn);
        return eff_1;
    }
    var reduceFn = function (acc, key) {
        var _a;
        return __assign({}, acc, (_a = {}, _a[key] = effectHandler.call(ctx, {
            effect: effects[key],
            promisify: promisify,
            cancel: cancel,
        }), _a));
    };
    var eff = Object.keys(effects).reduce(reduceFn, {});
    return eff;
}
exports.RACE = 'RACE';
function race(effects) {
    return {
        type: exports.RACE,
        effects: effects,
    };
}
exports.race = race;
var isRace = exports.typeDetector(exports.RACE);
function raceEffect(_a, promisify, genCancel) {
    var effects = _a.effects;
    var ctx = this;
    var forceCancel = noop;
    var cancel = new Promise(function (resolve) {
        if (genCancel && typeof genCancel.then === 'function') {
            genCancel.then(resolve);
        }
        forceCancel = resolve;
    });
    if (Array.isArray(effects)) {
        var mapFn = function (effect) {
            return promisify(effectHandler.call(ctx, { effect: effect, promisify: promisify, cancel: cancel }));
        };
        var eff = effects.map(mapFn);
        return Promise.race(eff)
            .then(function (result) {
            forceCancel('cancelled by race');
            return result;
        })
            .catch(function (err) {
            forceCancel(err);
        });
    }
    var keys = Object.keys(effects);
    return Promise.race(keys.map(function (key) {
        return promisify(effectHandler.call(ctx, { effect: effects[key], promisify: promisify, cancel: cancel })).then(function (result) {
            return {
                winner: key,
                result: result,
            };
        });
    })).then(function (result) {
        forceCancel('cancelled by race');
        return keys.reduce(function (acc, key) {
            if (result.winner === key) {
                acc[key] = result.result;
            }
            else {
                acc[key] = undefined;
            }
            return acc;
        }, {});
    });
}
exports.FORK = 'FORK';
exports.fork = function (fn) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    return ({
        type: exports.FORK,
        fn: fn,
        args: args,
    });
};
var isFork = exports.typeDetector(exports.FORK);
function forkEffect(_a, promisify, cancel) {
    var _this = this;
    var fn = _a.fn, args = _a.args;
    if (Array.isArray) {
        return speculation_1.default(function (resolve, reject) {
            var _a = fn, obj = _a[0], fnName = _a[1];
            promisify(obj[fnName].apply(obj, args))
                .then(noop)
                .catch(reject);
            resolve();
        }, cancel);
    }
    return speculation_1.default(function (resolve, reject) {
        var _a;
        promisify((_a = fn).call.apply(_a, [_this].concat(args)))
            .then(noop)
            .catch(reject);
        resolve();
    }, cancel);
}
exports.SPAWN = 'SPAWN';
exports.spawn = function (fn) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    return ({ type: exports.SPAWN, fn: fn, args: args });
};
var isSpawn = exports.typeDetector(exports.SPAWN);
function spawnEffect(_a, promisify) {
    var _this = this;
    var fn = _a.fn, args = _a.args;
    return speculation_1.default(function (resolve, reject, onCancel) {
        promisify(fn.call.apply(fn, [_this].concat(args)))
            .then(noop)
            .catch(reject);
        resolve();
        onCancel(function () {
            reject();
        });
    });
}
exports.DELAY = 'DELAY';
exports.delay = function (ms) { return ({
    type: exports.DELAY,
    ms: ms,
}); };
var isDelay = exports.typeDetector(exports.DELAY);
function delayEffect(_a, cancel) {
    var ms = _a.ms;
    return speculation_1.default(function (resolve, reject, onCancel) {
        var timerId = setTimeout(function () {
            resolve();
        }, ms);
        onCancel(function () {
            clearTimeout(timerId);
            reject('delay has been cancelled.');
        });
    }, cancel);
}
function effectHandler(_a) {
    var effect = _a.effect, promisify = _a.promisify, cancel = _a.cancel;
    var ctx = this;
    if (isCall(effect)) {
        return callEffect.call(ctx, effect, promisify, cancel);
    }
    if (isAll(effect)) {
        return allEffect.call(ctx, effect, promisify, cancel);
    }
    if (isRace(effect)) {
        return raceEffect.call(ctx, effect, promisify, cancel);
    }
    if (isSpawn(effect)) {
        return spawnEffect.call(ctx, effect, promisify);
    }
    if (isFork(effect)) {
        return forkEffect.call(ctx, effect, promisify, cancel);
    }
    if (isDelay(effect)) {
        return delayEffect.call(ctx, effect, cancel);
    }
    return effect;
}
exports.effectHandler = effectHandler;
function effectMiddleware(next) {
    return function (effect, promisify, cancel) {
        var nextEffect = effectHandler({ effect: effect, promisify: promisify, cancel: cancel });
        return next(nextEffect);
    };
}
exports.effectMiddleware = effectMiddleware;
//# sourceMappingURL=effect.js.map