"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var is_1 = require("./is");
var speculation_1 = require("./speculation");
var apply_middleware_1 = require("./apply-middleware");
function runtime(_a) {
    var fn = _a.fn, args = _a.args, cancel = _a.cancel, _b = _a.middleware, middleware = _b === void 0 ? [] : _b;
    var ctx = this;
    function promisify(obj, genCancel) {
        var nCancel = genCancel || cancel;
        if (!obj)
            return obj;
        if (is_1.isPromise(obj))
            return obj;
        if (typeof obj === 'function' ||
            is_1.isGeneratorFunction(obj) ||
            is_1.isGenerator(obj)) {
            return runtime.call(ctx, {
                fn: obj,
                args: [],
                cancel: nCancel,
                middleware: middleware,
            });
        }
        if (Array.isArray(obj)) {
            var eff = obj.map(function (o) { return promisify(o, nCancel); });
            return Promise.all(eff);
        }
        if (is_1.isObject(obj))
            return objectToPromise.call(ctx, obj);
        return obj;
    }
    function objectToPromise(obj) {
        var results = __assign({}, obj);
        var keys = Object.keys(obj);
        var promises = [];
        for (var i = 0; i < keys.length; i++) {
            var key = keys[i];
            var promise = promisify.call(ctx, obj[key]);
            if (promise && is_1.isPromise(promise)) {
                defer(promise, key);
            }
            else {
                results[key] = obj[key];
            }
        }
        return Promise.all(promises).then(function () { return results; });
        function defer(promise, key) {
            results[key] = undefined;
            promises.push(promise.then(function (res) {
                results[key] = res;
            }));
        }
    }
    return speculation_1.default(function (resolve, reject, onCancel) {
        var iter = typeof fn === 'function' ? fn.apply(ctx, args || []) : fn;
        if (!iter || typeof iter.next !== 'function') {
            return resolve(iter);
        }
        onCancel(onCancelled);
        onFulfilled();
        function onCancelled(error) {
            try {
                iter.throw(error);
                reject({ error: error });
            }
            catch (err) {
                return reject(err);
            }
        }
        function onFulfilled(res) {
            try {
                var ret = iter.next(res);
                next(ret);
                return null;
            }
            catch (e) {
                return reject(e);
            }
        }
        function onRejected(err) {
            try {
                var ret = iter.throw(err);
                next(ret);
            }
            catch (e) {
                return reject(e);
            }
        }
        function next(ret) {
            var value = ret.value;
            if (ret.done) {
                return resolve(value);
            }
            var taskValue = apply_middleware_1.default(middleware, ctx)(value, promisify, cancel);
            var promiseValue = promisify.call(ctx, taskValue);
            if (promiseValue && is_1.isPromise(promiseValue)) {
                return promiseValue.then(onFulfilled, onRejected);
            }
            var msg = "You may only yield a function, promise, generator, array, or" +
                (" object, but the following object was passed: \"" + ret.value + "\"");
            return onRejected(new TypeError(msg));
        }
    }, cancel);
}
exports.default = runtime;
//# sourceMappingURL=runtime.js.map